// Copyright (c) 2006 Nigel Tao.
// Licenced under the GNU General Public Licence (GPL) version 2.

#include "screen.h"

#include <libwnck/libwnck.h>
#include "string.h"

#include "window.h"
#include "workspace.h"

//------------------------------------------------------------------------------

#define NUMBER_OF_F_KEYS 12

static char *f_keys[] = {
	"F1", "F2", "F3", "F4", "F5", "F6",
	"F7", "F8", "F9", "F10", "F11", "F12"
};

//------------------------------------------------------------------------------

static guint active_window_changed_signal;
static guint active_workspace_changed_signal;
static guint window_closed_signal;
static guint window_opened_signal;
static guint workspace_created_signal;
static guint workspace_destroyed_signal;

//------------------------------------------------------------------------------

SSWorkspace *
ss_screen_get_nth_workspace (SSScreen *screen, int n)
{
	return (SSWorkspace *) g_list_nth_data (screen->workspaces, n);
}

//------------------------------------------------------------------------------

static void
update_search (SSScreen *screen, SSWindow *window, gchar** terms)
{
	gboolean matched;
	char* title;
	gchar* term;
	int t;
	
	matched = TRUE;
	title = (char *) wnck_window_get_name (window->wnck_window);
	title = g_ascii_strdown (title, strlen (title));
	
	t = 0;
	while (terms[t] != NULL) {
		term = terms[t];
		t++;
		
		if (strlen(term) == 0) {
			continue;
		}
		if (g_strrstr (title, term) == NULL) {
			matched = FALSE;
			break;
		}
	}
	
	ss_window_set_sensitive (window, matched);
	if (matched) {
		screen->num_search_matches++;
	}
	
	g_free (title);
}

//------------------------------------------------------------------------------

void
ss_screen_update_search (SSScreen *screen, const char *query)
{
	char *normalized_query;
	gchar** terms;
	SSWorkspace *workspace;
	SSWindow *window;
	GList *i;
	GList *j;
	
	normalized_query = g_ascii_strdown (query, strlen (query));
	terms = g_strsplit (normalized_query, " ", 0);
	screen->num_search_matches = 0;
	
	for (i = screen->workspaces; i; i = i->next) {
		workspace = (SSWorkspace *) i->data;
		for (j = workspace->windows; j; j = j->next) {
			window = (SSWindow *) j->data;
			
			update_search (screen, window, terms);
		}
	}
	
	g_strfreev (terms);
	g_free (normalized_query);
}

//------------------------------------------------------------------------------

void
ss_screen_activate_next_window (SSScreen *screen, gboolean backwards, guint32 time)
{
	SSWorkspace *workspace;
	SSWindow *window;
	GList *i;
	GList *j;
	
	SSWindow *first_sensitive_window;
	SSWindow *previous_sensitive_window;
	gboolean should_activate_last_sensitive_window;
	gboolean should_activate_next_sensitive_window;
	
	if (screen->active_window == NULL) {
		return;
	}

	first_sensitive_window    = NULL;
	previous_sensitive_window = NULL;
	should_activate_last_sensitive_window = FALSE;
	should_activate_next_sensitive_window = FALSE;
	
	for (i = screen->workspaces; i; i = i->next) {
		workspace = (SSWorkspace *) i->data;
		for (j = workspace->windows; j; j = j->next) {
			window = (SSWindow *) j->data;
			
			// We've found the active window, and we're searching backwards.
			if ((backwards) && (window == screen->active_window)) {
				if (previous_sensitive_window == NULL) {
					should_activate_last_sensitive_window = TRUE;
				} else {
					ss_window_activate_workspace_and_window (previous_sensitive_window, time);
					return;
				}
			}
			
			if (window->sensitive) {
				if (should_activate_next_sensitive_window) {
					ss_window_activate_workspace_and_window (window, time);
					return;
				}
				
				previous_sensitive_window = window;
				
				if (first_sensitive_window == NULL) {
					first_sensitive_window = window;
				}
			}
			
			// We've found the active window, and we're searching forwards.
			if ((! backwards) && (window == screen->active_window)) {
				should_activate_next_sensitive_window = TRUE;
			}
		}
	}
	
	if (should_activate_next_sensitive_window) {
		ss_window_activate_workspace_and_window (first_sensitive_window, time);
		return;
	}
	
	if (should_activate_last_sensitive_window) {
		ss_window_activate_workspace_and_window (previous_sensitive_window, time);
		return;
	}
}

//------------------------------------------------------------------------------

void
ss_screen_activate_next_window_in_stacking_order (SSScreen *screen, gboolean backwards, guint32 time)
{
	WnckWindow *active_wnck_window;
	WnckWindow *wnck_window;
	gboolean ww_is_aww;
	int active_workspace_id;
	int workspace_id;
	GList *i;
	
	WnckWindow *first_eligible_wnck_window;
	WnckWindow *previous_eligible_wnck_window;
	gboolean should_activate_last_eligible_wnck_window;
	gboolean should_activate_next_eligible_wnck_window;
	
	if (screen->active_window == NULL) {
		return;
	}
	
	// TODO - this is probably an ugly hack that's exposed to a potential
	// race condition.  A better way to go about this would be to suppress
	// the window_stacking_change signal, I suppose.
	// For example, if a window gets opened in between setting this flag
	// and in the event handler on_window_stacking_changed, then the
	// window_stacking_order list will be copied at the wrong point,
	// possibly.
	screen->should_ignore_next_window_stacking_change = TRUE;
	
	first_eligible_wnck_window    = NULL;
	previous_eligible_wnck_window = NULL;
	should_activate_last_eligible_wnck_window = FALSE;
	should_activate_next_eligible_wnck_window = FALSE;

	active_wnck_window = screen->active_window->wnck_window;
	active_workspace_id = wnck_workspace_get_number (wnck_window_get_workspace (active_wnck_window));
	
	// Note that wnck_windows_in_stacking_order is in bottom-to-top
	// order, so that if we're searching forwards (just like Alt-Tab)
	// then we want the previous_eligible_wnck_window.  This is the
	// opposite ordering than in ss_screen_activate_next_window.
	for (i = screen->wnck_windows_in_stacking_order; i; i = i->next) {
		wnck_window = (WnckWindow *) i->data;
		workspace_id = wnck_workspace_get_number (wnck_window_get_workspace (wnck_window));
		
		if (workspace_id != active_workspace_id) {
			continue;
		}
		
		ww_is_aww = (wnck_window_get_xid (wnck_window) ==
			wnck_window_get_xid (active_wnck_window));

		// We've found the active window, and we're searching forwards.
		if ((! backwards) && (ww_is_aww)) {
			if (previous_eligible_wnck_window == NULL) {
				should_activate_last_eligible_wnck_window = TRUE;
			} else {
				wnck_window_activate (previous_eligible_wnck_window, time);
				return;
			}
		}
		
		if (should_activate_next_eligible_wnck_window) {
			wnck_window_activate (wnck_window, time);
			return;
		}
		
		previous_eligible_wnck_window = wnck_window;
		
		if (first_eligible_wnck_window == NULL) {
			first_eligible_wnck_window = wnck_window;
		}
		
		// We've found the active window, and we're searching backwards.
		if ((backwards) && (ww_is_aww)) {
			should_activate_next_eligible_wnck_window = TRUE;
		}
	}
	
	if (should_activate_next_eligible_wnck_window) {
		wnck_window_activate (first_eligible_wnck_window, time);
		return;
	}
	
	if (should_activate_last_eligible_wnck_window) {
		wnck_window_activate (previous_eligible_wnck_window, time);
		return;
	}
}

//------------------------------------------------------------------------------

static SSWindow *
get_ss_window_from_wnck_window (SSScreen *screen, WnckWindow *wnck_window)
{
	gulong xid;
	SSWorkspace *workspace;
	SSWindow *window;
	GList *i;
	GList *j;
	
	if (wnck_window == NULL) {
		return NULL;
	}
	
	xid = wnck_window_get_xid (wnck_window);
	
	// Walk through the list of list of windows, matching by X ID.
	for (i = screen->workspaces; i; i = i->next) {
		workspace = (SSWorkspace *) i->data;
		for (j = workspace->windows; j; j = j->next) {
			window = (SSWindow *) j->data;
			if (xid == wnck_window_get_xid (window->wnck_window)) {
				return window;
			}
		}
	}
	
	return NULL;
}

//------------------------------------------------------------------------------

static SSWorkspace *
get_ss_workspace_from_wnck_workspace (SSScreen *screen, WnckWorkspace *wnck_workspace)
{
	SSWorkspace *workspace;
	GList *i;
	
	if (wnck_workspace == NULL) {
		return NULL;
	}
	
	for (i = screen->workspaces; i; i = i->next) {
		workspace = (SSWorkspace *) i->data;
		if (wnck_workspace == workspace->wnck_workspace) {
			return workspace;
		}
	}
	
	return NULL;
}

//------------------------------------------------------------------------------

static void
update_workspace_titles (SSScreen *screen)
{
	SSWorkspace *workspace;
	GList *i;
	int j;
	
	for (j = 0, i = screen->workspaces; i; j++, i = i->next) {
		workspace = (SSWorkspace *) i->data;
		workspace->title = (j < NUMBER_OF_F_KEYS) ? f_keys[j] : "";
	}
}

//------------------------------------------------------------------------------

static void
update_for_active_window (SSScreen *screen)
{
	WnckWindow *wnck_window;
	wnck_window = wnck_screen_get_active_window (screen->wnck_screen);
	screen->active_window = get_ss_window_from_wnck_window (screen, wnck_window);
}

//------------------------------------------------------------------------------

static void
update_for_active_workspace (SSScreen *screen)
{
	screen->active_workspace_id = wnck_workspace_get_number
		(wnck_screen_get_active_workspace (screen->wnck_screen));
	screen->active_workspace = ss_screen_get_nth_workspace
		(screen, screen->active_workspace_id);
}

//------------------------------------------------------------------------------

static SSWorkspace *
add_workspace_to_screen (SSScreen *screen, WnckWorkspace *wnck_workspace)
{
	SSWorkspace *workspace;
	workspace = ss_workspace_new (screen, wnck_workspace);
	screen->workspaces = g_list_append (screen->workspaces, workspace);
	return workspace;
}

//------------------------------------------------------------------------------

static SSWindow *
add_window_to_screen (SSScreen *screen, WnckWindow *wnck_window)
{
	SSWindow *window;
	SSWorkspace *workspace;
	int n;
	
	if (wnck_window_is_skip_pager (wnck_window)) {
		return NULL;
	}
	
	n = wnck_workspace_get_number (wnck_window_get_workspace (wnck_window));
	workspace = ss_screen_get_nth_workspace (screen, n);
	
	window = ss_window_new (workspace, wnck_window);
	if (wnck_window_is_active (wnck_window)) {
		screen->active_window = window;
	}
	ss_workspace_add_window (workspace, window);
	return window;
}

//------------------------------------------------------------------------------

static void
update_window_label_width (SSScreen *screen)
{
	PangoContext *context;
	PangoFontMetrics *metrics;
	int width, char_width;
	SSWorkspace *workspace;
	SSWindow *window;
	GList *i;
	GList *j;
	
	context = gtk_widget_get_pango_context (screen->widget_for_font_metrics);
	metrics = pango_context_get_metrics (context,
		screen->widget_for_font_metrics->style->font_desc, NULL);
	char_width = PANGO_PIXELS (pango_font_metrics_get_approximate_char_width (metrics));
	pango_font_metrics_unref (metrics);
	
	// The popup should, at most, be approximately two thirds of the screen
	// width.  This is completely arbitrary, but it looks OK on my machine.
	width = (screen->screen_width * 2 / 3) / screen->num_workspaces;
	// Subtract off a bit for the icon, and the remainder is for the label.
	width -= 30;
	// convert from pixels to chars.
	width /= char_width;
	
	screen->label_max_width_chars = width;
	
	for (i = screen->workspaces; i; i = i->next) {
		workspace = (SSWorkspace *) i->data;
		for (j = workspace->windows; j; j = j->next) {
			window = (SSWindow *) j->data;
			ss_window_update_label_max_width_chars (window);
		}
	}
}

//------------------------------------------------------------------------------

static void
on_active_window_changed (WnckScreen *wnck_screen, gpointer data)
{
	SSScreen *screen;
	screen = (SSScreen *) data;
	update_for_active_window (screen);
	g_signal_emit (screen, active_window_changed_signal, 0, NULL);
}

//------------------------------------------------------------------------------

static void
on_active_workspace_changed (WnckScreen *wnck_screen, gpointer data)
{
	SSScreen *screen;
	screen = (SSScreen *) data;
	update_for_active_workspace (screen);
	g_signal_emit (screen, active_workspace_changed_signal, 0, NULL);
}

//------------------------------------------------------------------------------

static void
on_window_closed (WnckScreen *wnck_screen, WnckWindow *wnck_window, gpointer data)
{
	SSScreen *screen;
	SSWindow *window;
	int n;
	
	screen = (SSScreen *) data;
	window = get_ss_window_from_wnck_window (screen, wnck_window);
	n = wnck_workspace_get_number (wnck_window_get_workspace (wnck_window));
	ss_workspace_remove_window (ss_screen_get_nth_workspace (screen, n), window);
	
	g_signal_emit (screen, window_closed_signal, 0, window);
	ss_window_free (window);
}

//------------------------------------------------------------------------------

static void
on_window_opened (WnckScreen *wnck_screen, WnckWindow *wnck_window, gpointer data)
{
	SSScreen *screen;
	SSWindow *window;
	
	screen = (SSScreen *) data;
	window = add_window_to_screen (screen, wnck_window);
	
	g_signal_emit (screen, window_opened_signal, 0, window);
}

//------------------------------------------------------------------------------

void
ss_screen_update_wnck_windows_in_stacking_order (SSScreen *screen)
{
	WnckWindow *wnck_window;
	GList *i;
	
	if (screen->wnck_windows_in_stacking_order != NULL) {
		g_list_free (screen->wnck_windows_in_stacking_order);
		screen->wnck_windows_in_stacking_order = NULL;
	}
	
	// Make a copy of wnck_screen_get_windows_stacked (screen->wnck_screen),
	// filtering out those windows that we do not show.
	for (i = wnck_screen_get_windows_stacked (screen->wnck_screen); i; i = i->next) {
		wnck_window = (WnckWindow *) i->data;
		if (wnck_window_is_skip_pager (wnck_window)) {
			continue;
		}
		
		screen->wnck_windows_in_stacking_order = g_list_append
			(screen->wnck_windows_in_stacking_order, wnck_window);
	}
	
	for (i = screen->wnck_windows_in_stacking_order; i; i = i->next) {
		wnck_window = (WnckWindow *) i->data;
	}
}

//------------------------------------------------------------------------------

static void
on_window_stacking_changed (WnckScreen *wnck_screen, gpointer data)
{
	SSScreen *screen;
	screen = (SSScreen *) data;
	
	if (screen->should_ignore_next_window_stacking_change) {
		screen->should_ignore_next_window_stacking_change = FALSE;
		return;
	}
	
	ss_screen_update_wnck_windows_in_stacking_order (screen);
}

//------------------------------------------------------------------------------

static void
on_workspace_created (WnckScreen *wnck_screen, WnckWorkspace *wnck_workspace, gpointer data)
{
	SSScreen *screen;
	SSWorkspace *workspace;
	
	screen = (SSScreen *) data;
	screen->num_workspaces = wnck_screen_get_workspace_count (wnck_screen);
	workspace = add_workspace_to_screen (screen, wnck_workspace);
	
	update_window_label_width (screen);
	update_workspace_titles (screen);
	g_signal_emit (screen, workspace_created_signal, 0, workspace);
}

//------------------------------------------------------------------------------

static void
on_workspace_destroyed (WnckScreen *wnck_screen, WnckWorkspace *wnck_workspace, gpointer data)
{
	SSScreen *screen;
	SSWorkspace *workspace;

	screen = (SSScreen *) data;
	screen->num_workspaces -= 1;
	workspace = get_ss_workspace_from_wnck_workspace (screen, wnck_workspace);
	screen->workspaces = g_list_remove (screen->workspaces, workspace);
	
	update_window_label_width (screen);
	update_workspace_titles (screen);
	g_signal_emit (screen, workspace_destroyed_signal, 0, workspace);
	ss_workspace_free (workspace);
}

//------------------------------------------------------------------------------

static void
ss_screen_class_init (SSScreenClass *klass)
{
	GType wcs_types[1] = { G_TYPE_POINTER };
	
	active_window_changed_signal = g_signal_newv (
		"active-window-changed",
		G_TYPE_FROM_CLASS (klass),
		G_SIGNAL_RUN_LAST,
		NULL, NULL, NULL,
		g_cclosure_marshal_VOID__VOID,
		G_TYPE_NONE,
		0, NULL);
	
	active_workspace_changed_signal = g_signal_newv (
		"active-workspace-changed",
		G_TYPE_FROM_CLASS (klass),
		G_SIGNAL_RUN_LAST,
		NULL, NULL, NULL,
		g_cclosure_marshal_VOID__VOID,
		G_TYPE_NONE,
		0, NULL);
	
	window_closed_signal = g_signal_newv (
		"window_closed",
		G_TYPE_FROM_CLASS (klass),
		G_SIGNAL_RUN_LAST,
		NULL, NULL, NULL,
		g_cclosure_marshal_VOID__POINTER,
		G_TYPE_NONE,
		1, wcs_types);
	
	window_opened_signal = g_signal_newv (
		"window_opened",
		G_TYPE_FROM_CLASS (klass),
		G_SIGNAL_RUN_LAST,
		NULL, NULL, NULL,
		g_cclosure_marshal_VOID__POINTER,
		G_TYPE_NONE,
		1, wcs_types);
	
	workspace_created_signal = g_signal_newv (
		"workspace_created",
		G_TYPE_FROM_CLASS (klass),
		G_SIGNAL_RUN_LAST,
		NULL, NULL, NULL,
		g_cclosure_marshal_VOID__POINTER,
		G_TYPE_NONE,
		1, wcs_types);
	
	workspace_destroyed_signal = g_signal_newv (
		"workspace_destroyed",
		G_TYPE_FROM_CLASS (klass),
		G_SIGNAL_RUN_LAST,
		NULL, NULL, NULL,
		g_cclosure_marshal_VOID__POINTER,
		G_TYPE_NONE,
		1, wcs_types);
}

//------------------------------------------------------------------------------

GType
ss_screen_get_type (void)
{
	static GType type = 0;
	if (type == 0) {
		static const GTypeInfo info = {
			sizeof (SSScreenClass),
			NULL,   // base_init
			NULL,   // base_finalize
			(GClassInitFunc) ss_screen_class_init,
			NULL,   // class_finalize
			NULL,   // class_data
			sizeof (SSScreen),
			0,      // n_preallocs
			NULL    // instance_init
		};
		type = g_type_register_static (G_TYPE_OBJECT, "SSScreenType", &info, 0);
	}
	return type;
}

//------------------------------------------------------------------------------

SSScreen *
ss_screen_new (WnckScreen *wnck_screen)
{
	SSScreen *screen;
	
	GList *wnck_windows;
	int i;
	
	wnck_screen_force_update (wnck_screen);
	
	screen = (SSScreen *) g_object_new (SS_TYPE_SCREEN, NULL);
	screen->wnck_screen = wnck_screen;
	screen->screen_width  = wnck_screen_get_width (wnck_screen);
	screen->screen_height = wnck_screen_get_height (wnck_screen);
	screen->screen_aspect = (double) screen->screen_height / (double) screen->screen_width;
	
	screen->num_workspaces = wnck_screen_get_workspace_count (wnck_screen);
	
	screen->active_window = NULL;
	screen->active_workspace = NULL;
	screen->active_workspace_id = -1;
	
	screen->wnck_windows_in_stacking_order = NULL;
	screen->should_ignore_next_window_stacking_change = FALSE;
	ss_screen_update_wnck_windows_in_stacking_order (screen);
	
	screen->num_search_matches = 0;
	
	screen->widget_for_font_metrics = gtk_label_new (NULL);
	screen->label_max_width_chars = 256;
	update_window_label_width (screen);
	
	// Add existing workspaces, and then existing windows
	screen->workspaces = NULL;
	for (i = 0; i < screen->num_workspaces; i++) {
		add_workspace_to_screen (screen, wnck_screen_get_workspace (wnck_screen, i));
	}
	update_workspace_titles (screen);
	
	wnck_windows = wnck_screen_get_windows (wnck_screen);
	for (; wnck_windows; wnck_windows = wnck_windows->next) {
		add_window_to_screen (screen, WNCK_WINDOW (wnck_windows->data));
	}
	
	// Listen for new workspaces, and new windows
	g_signal_connect (G_OBJECT (wnck_screen), "active_window_changed",
		G_CALLBACK (on_active_window_changed),
		screen);
	g_signal_connect (G_OBJECT (wnck_screen), "active_workspace_changed",
		G_CALLBACK (on_active_workspace_changed),
		screen);
	g_signal_connect (G_OBJECT (wnck_screen), "window_closed",
		G_CALLBACK (on_window_closed),
		screen);
	g_signal_connect (G_OBJECT (wnck_screen), "window_opened",
		G_CALLBACK (on_window_opened),
		screen);
	g_signal_connect (G_OBJECT (wnck_screen), "window_stacking_changed",
		G_CALLBACK (on_window_stacking_changed),
		screen);
	g_signal_connect (G_OBJECT (wnck_screen), "workspace_created",
		G_CALLBACK (on_workspace_created),
		screen);
	g_signal_connect (G_OBJECT (wnck_screen), "workspace_destroyed",
		G_CALLBACK (on_workspace_destroyed),
		screen);
	
	update_for_active_workspace (screen);
	
	return screen;
}
