// Copyright (c) 2006 Nigel Tao.
// Licenced under the GNU General Public Licence (GPL) version 2.

#include "window.h"

#include "screen.h"
#include "workspace.h"

//------------------------------------------------------------------------------

void
ss_window_update_label_max_width_chars (SSWindow *window)
{
	gtk_label_set_max_width_chars (GTK_LABEL (window->label), window->workspace->screen->label_max_width_chars);
}

//------------------------------------------------------------------------------

void
ss_window_set_sensitive (SSWindow *window, gboolean sensitive)
{
	gtk_widget_set_sensitive (GTK_WIDGET (window->image), sensitive);
	gtk_widget_set_sensitive (GTK_WIDGET (window->label), sensitive);
	window->sensitive = sensitive;
}

//------------------------------------------------------------------------------

void
ss_window_activate_workspace_and_window (SSWindow *window, guint32 time)
{
	if (window == NULL) {
		return;
	}
	
	wnck_workspace_activate (window->workspace->wnck_workspace, time);
	wnck_window_activate (window->wnck_window, time + 1);
}

//------------------------------------------------------------------------------

static gboolean
on_button_press_event (GtkWidget *widget, GdkEventButton *event, gpointer data)
{
	ss_window_activate_workspace_and_window ((SSWindow *) data, event->time);
	return TRUE;
}

//------------------------------------------------------------------------------

static void
on_geometry_changed (WnckWindow *wnck_window, gpointer data)
{
	SSWindow *window;
	window = (SSWindow *) data;
	gtk_widget_queue_draw (window->workspace->header);
}

//------------------------------------------------------------------------------

static void
on_icon_changed (WnckWindow *wnck_window, gpointer data)
{
	SSWindow *window;
	window = (SSWindow *) data;
	gtk_image_set_from_pixbuf (GTK_IMAGE(window->image), wnck_window_get_mini_icon (wnck_window));
	gtk_widget_queue_draw (window->workspace->header);
}


//------------------------------------------------------------------------------

static void
on_name_changed (WnckWindow *wnck_window, gpointer data)
{
	SSWindow *window;
	window = (SSWindow *) data;
	gtk_label_set_text (GTK_LABEL (window->label),
		wnck_window_get_name (window->wnck_window));
	gtk_widget_queue_draw (window->workspace->header);
}

//------------------------------------------------------------------------------

static void
on_workspace_changed (WnckWindow *wnck_window, gpointer data)
{
	SSWindow *window;
	SSWorkspace *old_workspace;
	SSWorkspace *new_workspace;
	int new_workspace_id;
	
	window = (SSWindow *) data;
	old_workspace = window->workspace;
	new_workspace_id = wnck_workspace_get_number (wnck_window_get_workspace (wnck_window));
	new_workspace = ss_screen_get_nth_workspace (old_workspace->screen, new_workspace_id);
	
	ss_workspace_remove_window (old_workspace, window);
	window->workspace = new_workspace;
	ss_workspace_add_window (new_workspace, window);
}

//------------------------------------------------------------------------------

SSWindow *
ss_window_new (SSWorkspace *workspace, WnckWindow *wnck_window)
{
	SSWindow *w;
	GtkWidget *eventbox;
	GtkWidget *hbox;
	GtkWidget *image;
	GtkWidget *label;
	
	eventbox = gtk_event_box_new ();
	gtk_event_box_set_visible_window (GTK_EVENT_BOX (eventbox), FALSE);
	
	hbox = gtk_hbox_new (FALSE, 3);
	gtk_container_add (GTK_CONTAINER (eventbox), hbox);
	
	// Spacer, in order to keep the image centered in the column.
	gtk_box_pack_start (GTK_BOX (hbox), gtk_label_new (NULL), FALSE, FALSE, 0);
	
	image = gtk_image_new ();
	gtk_image_set_from_pixbuf (GTK_IMAGE(image), wnck_window_get_mini_icon (wnck_window));
	gtk_box_pack_start (GTK_BOX (hbox), image, FALSE, FALSE, 0);
	
	label = gtk_label_new (wnck_window_get_name (wnck_window));
	gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
	gtk_label_set_max_width_chars (GTK_LABEL (label), workspace->screen->label_max_width_chars);
	gtk_label_set_ellipsize (GTK_LABEL (label), PANGO_ELLIPSIZE_MIDDLE);
	gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
	
	w = g_new (SSWindow, 1);
	w->workspace = workspace;
	w->wnck_window = wnck_window;
	w->widget = eventbox;
	w->image = image;
	w->label = label;
	w->sensitive = TRUE;
	w->signal_id_geometry_changed =
		g_signal_connect (G_OBJECT (wnck_window), "geometry-changed",
		(GCallback) on_geometry_changed,
		w);
	w->signal_id_icon_changed =
		g_signal_connect (G_OBJECT (wnck_window), "icon-changed",
		(GCallback) on_icon_changed,
		w);
	w->signal_id_name_changed =
		g_signal_connect (G_OBJECT (wnck_window), "name-changed",
		(GCallback) on_name_changed,
		w);
	w->signal_id_workspace_changed =
		g_signal_connect (G_OBJECT (wnck_window), "workspace-changed",
		(GCallback) on_workspace_changed,
		w);
	g_signal_connect (G_OBJECT (eventbox), "button-press-event",
		(GCallback) on_button_press_event,
		w);
	g_object_ref (w->widget);
	return w;
}

//------------------------------------------------------------------------------

void
ss_window_free (SSWindow *window)
{
	g_signal_handler_disconnect (G_OBJECT (window->wnck_window),
		window->signal_id_geometry_changed);
	g_signal_handler_disconnect (G_OBJECT (window->wnck_window),
		window->signal_id_icon_changed);
	g_signal_handler_disconnect (G_OBJECT (window->wnck_window),
		window->signal_id_name_changed);
	g_signal_handler_disconnect (G_OBJECT (window->wnck_window),
		window->signal_id_workspace_changed);
	g_object_unref (window->widget);
	g_free (window);
}
