// Copyright (c) 2006 Nigel Tao.
// Licenced under the GNU General Public Licence (GPL) version 2.

#include "workspace.h"

#include "screen.h"
#include "window.h"

//------------------------------------------------------------------------------

#define MINI_WORKSPACE_FONT_HEIGHT 16
#define MINI_WORKSPACE_WIDTH 48

//------------------------------------------------------------------------------

static void
workspace_remove_window_widgets (SSWorkspace *workspace)
{
	GList *children;
	GList *i;
	GtkWidget *child;
	GtkContainer *container;
	
	container = GTK_CONTAINER (workspace->window_container);
	children = gtk_container_get_children (container);
	for (i = children; i; i = i->next) {
		child = GTK_WIDGET (i->data);
		gtk_container_remove (container, child);
	}
	g_list_free (children);
}

//------------------------------------------------------------------------------

static void
workspace_add_window_widgets (SSWorkspace *workspace)
{
	GList *i;
	SSWindow *window;
	
	for (i = workspace->windows; i; i = i->next) {
		window = (SSWindow *) i->data;
		gtk_box_pack_start (GTK_BOX (workspace->window_container),
			window->widget, FALSE, FALSE, 0);
	}
}

//------------------------------------------------------------------------------

void
ss_workspace_add_window (SSWorkspace *workspace, SSWindow *window)
{
	workspace->windows = g_list_append (workspace->windows, window);
	gtk_box_pack_start (GTK_BOX (workspace->window_container), window->widget, TRUE, TRUE, 0);
}

//------------------------------------------------------------------------------

void
ss_workspace_remove_window (SSWorkspace *workspace, SSWindow *window)
{
	workspace->windows = g_list_remove (workspace->windows, window);
	gtk_container_remove (GTK_CONTAINER (workspace->window_container), window->widget);
}

//------------------------------------------------------------------------------

void
ss_workspace_reorder_window (SSWorkspace *workspace, SSWindow *window, int new_index)
{
	workspace->windows = g_list_remove (workspace->windows, window);
	workspace->windows = g_list_insert (workspace->windows, window, new_index);
	workspace_remove_window_widgets (workspace);
	workspace_add_window_widgets (workspace);
}

//------------------------------------------------------------------------------

static gboolean
on_button_press_event (GtkWidget *widget, GdkEventButton *event, gpointer data)
{
	wnck_workspace_activate (((SSWorkspace *) data)->wnck_workspace, event->time);
	return TRUE;
}

//------------------------------------------------------------------------------

static void
on_expose_event_draw_text (GtkWidget *widget, SSWorkspace *workspace)
{
	cairo_t *c;
	cairo_text_extents_t extents;
	int x, y;
	
	c = gdk_cairo_create (widget->window);
	cairo_select_font_face (c, "Sans", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_BOLD);
	cairo_set_font_size (c, MINI_WORKSPACE_FONT_HEIGHT);
	
	cairo_text_extents (c, workspace->title, &extents);
	x = (widget->allocation.width  - extents.width)  / 2;
	y = (widget->allocation.height + extents.height) / 2;
	
	cairo_set_source_rgba (c, 0, 0, 0, 0.25);
	cairo_move_to (c, x-1, y);
	cairo_show_text (c, workspace->title);
	cairo_set_source_rgba (c, 1, 1, 1, 0.25);
	cairo_move_to (c, x, y+1);
	cairo_show_text (c, workspace->title);
}

//------------------------------------------------------------------------------

static gboolean
on_expose_event (GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
	SSWorkspace *workspace;
	int screen_width, screen_height;
	double width_ratio, height_ratio;
	int x, y, w, h;
	GList *i;
	SSWindow *active_window;
	WnckWindow *wnck_window;
	int state;
	GdkRectangle r;
	
	workspace = (SSWorkspace *) data;
	screen_width  = workspace->screen->screen_width;
	screen_height = workspace->screen->screen_height;
	active_window = workspace->screen->active_window;
	
	x = widget->allocation.x;
	y = widget->allocation.y;
	w = widget->allocation.width;
	h = widget->allocation.height;
	
	state = (workspace == workspace->screen->active_workspace) ? GTK_STATE_SELECTED : GTK_STATE_NORMAL;
	gdk_draw_rectangle (widget->window, 
		widget->style->dark_gc[state], TRUE,
		1, 1, w-2, h-2);
	gdk_draw_rectangle (widget->window, 
		widget->style->base_gc[state], FALSE,
		0, 0, w-1, h-1);
	
	width_ratio  = (double) w / (double) screen_width;
	height_ratio = (double) h / (double) screen_height;
	
	for (i = workspace->screen->wnck_windows_in_stacking_order; i; i = i->next) {
		wnck_window = (WnckWindow *) i->data;
		if (wnck_window_get_workspace (wnck_window) != workspace->wnck_workspace) {
			continue;
		}
		if (wnck_window_is_minimized (wnck_window)) {
			continue;
		}
		
		state = GTK_STATE_ACTIVE;
		if ((active_window != NULL) &&
			((wnck_window_get_xid (wnck_window) == wnck_window_get_xid (active_window->wnck_window)))) {
			state = GTK_STATE_SELECTED;
		}
		wnck_window_get_geometry (wnck_window, &r.x, &r.y, &r.width, &r.height);
		
		r.x = 0.5 + r.x * width_ratio;
		r.y = 0.5 + r.y * height_ratio;
		r.width  = 0.5 + r.width  * width_ratio;
		r.height = 0.5 + r.height * height_ratio;
		
		if (r.width < 3) {
			r.width = 3;
		}
		if (r.height < 3) {
			r.height = 3;
		}
		
		gdk_draw_rectangle (widget->window, 
			widget->style->bg_gc[state], TRUE,
			r.x+1, r.y+1, r.width-2, r.height-2);
		gdk_draw_rectangle (widget->window, 
			widget->style->fg_gc[state], FALSE,
			r.x,   r.y,   r.width-1, r.height-1);
	}
	
	on_expose_event_draw_text (widget, workspace);
	return FALSE;
}

//------------------------------------------------------------------------------

SSWorkspace *
ss_workspace_new (SSScreen *screen, WnckWorkspace *wnck_workspace)
{
	SSWorkspace *w;
	GtkWidget *box;
	GtkWidget *align;
	GtkWidget *header;
	GtkWidget *align_h;
	GtkWidget *box_2;
	GtkWidget *align_2;
	
	box = gtk_vbox_new (FALSE, 3);
	align = gtk_alignment_new (0.5, 0.0, 0.0, 0.0);
	gtk_container_add (GTK_CONTAINER (align), box);
	
	header = gtk_drawing_area_new ();
	gtk_widget_add_events (header, GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK);
	gtk_widget_set_size_request (header, MINI_WORKSPACE_WIDTH,
		MINI_WORKSPACE_WIDTH * screen->screen_aspect);
	
	align_h = gtk_alignment_new (0.5, 0.5, 0.0, 0.0);
	gtk_container_add (GTK_CONTAINER (align_h), header);
	gtk_box_pack_start (GTK_BOX (box), align_h, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX (box), gtk_hseparator_new (), TRUE, TRUE, 0);
	
	box_2 = gtk_vbox_new (FALSE, 6);
	align_2 = gtk_alignment_new (0.5, 0.5, 0.0, 0.0);
	gtk_container_add (GTK_CONTAINER (align_2), box_2);
	gtk_box_pack_start (GTK_BOX (box), align_2, TRUE, TRUE, 0);
	
	w = g_new (SSWorkspace, 1);
	w->screen = screen;
	w->wnck_workspace = wnck_workspace;
	w->widget = align;
	w->header = header;
	w->window_container = box_2;
	w->title = "";
	w->windows = NULL;
	g_signal_connect (G_OBJECT (header), "expose-event",
		(GCallback) on_expose_event,
		w);
	g_signal_connect (G_OBJECT (header), "button-press-event",
		(GCallback) on_button_press_event,
		w);
	g_object_ref (w->widget);
	return w;
}

//------------------------------------------------------------------------------

void
ss_workspace_free (SSWorkspace *workspace)
{
	g_list_free (workspace->windows);
	g_object_unref (workspace->widget);
	g_free (workspace);
}
